# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import re

from typing import List

import edge_tts

from edge_tts import VoicesManager, list_voices
from edge_tts.exceptions import NoAudioReceived
from iso639 import Lang

from open_dubbing.text_to_speech import TextToSpeech, Voice


class TextToSpeechEdge(TextToSpeech):

    async def _create_manager(self):
        voice_manager = await VoicesManager.create()
        return voice_manager

    def __init__(self, device="cpu"):
        super().__init__()
        self.device = device

    def get_available_voices(self, language_code: str) -> List[Voice]:
        voices = []
        iso_639_1 = self._get_iso_639_1(language_code)

        voice_manager = asyncio.run(self._create_manager())

        edge_voices = voice_manager.find(Language=iso_639_1)
        for edge_voice in edge_voices:
            if not all(key in edge_voice for key in ["ShortName", "Gender", "Locale"]):
                logging.warning(
                    f"Skipping voice '{edge_voice}' since is missing some fields"
                )
                continue

            voice = Voice(
                name=edge_voice["ShortName"],
                gender=edge_voice["Gender"],
                region=edge_voice["Locale"],
            )
            voices.append(voice)
            logging.debug(
                f'shortname: {edge_voice["ShortName"]}, gender: {edge_voice["Gender"]}, locale: {edge_voice["Locale"]}'
            )

        logging.debug(
            f"text_to_speech_edge.get_available_voices: {voices} for language {language_code}"
        )

        return voices

    def _get_iso_639_1(self, iso_639_3: str):
        o = Lang(iso_639_3)
        iso_639_1 = o.pt1
        return iso_639_1

    def _does_voice_supports_speeds(self):
        return True

    async def _save(self, text, speed, assigned_voice, output_filename):
        per = (100 * speed) - 100
        str_per = f"+{per:0.0f}%"
        max_retries = 3
        for attempt in range(1, max_retries + 1):
            try:
                communicate = edge_tts.Communicate(text, assigned_voice, rate=str_per)
                await communicate.save(output_filename)
                return
            except NoAudioReceived:
                if attempt == max_retries:
                    logging.error(
                        "text_to_speech_edge._save. Max retries reached. Could not save audio."
                    )
                    raise
                else:
                    logging.warning(
                        f"text_to_speech_edge._save. No audio received, retrying attempt {attempt}."
                    )
                    await asyncio.sleep(30)

    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        speed: float,
    ) -> str:

        asyncio.run(self._save(text, speed, assigned_voice, output_filename))
        logging.debug(
            f"text_to_speech.client_edge.synthesize_speech: assigned_voice: {assigned_voice}, output_filename: '{output_filename}'"
        )
        return output_filename

    async def _get_list_voices(self):
        return await list_voices()

    def get_languages(self):
        voices = asyncio.run(self._get_list_voices())

        pattern = r"^([a-z]{2})-(.*)$"
        locales = set()
        for voice in voices:
            locale = voice["Locale"]
            _match = re.search(pattern, locale)
            if not _match:
                continue

            iso_639_1 = _match.group(1)
            o = Lang(iso_639_1)
            iso_639_3 = o.pt3

            locales.add(iso_639_3)
        languages = sorted(locales)
        logging.debug(f"text_to_speech.client_edge.get_languages: {languages}'")
        return languages



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import tempfile

from typing import List
from urllib.parse import urljoin

import requests

from open_dubbing.text_to_speech import TextToSpeech, Voice


class TextToSpeechAPI(TextToSpeech):

    def __init__(self, device="cpu", server=""):
        super().__init__()
        self.server = server
        self.device = device
        self.voices = None

    def _get_voices(self):
        if not self.voices:
            url = urljoin(self.server, "/voices")
            response = requests.get(url)
            self.voices = response.json()

        return self.voices

    def get_available_voices(self, language_code: str) -> List[Voice]:
        voices = []

        for server_voice in self._get_voices():
            language = server_voice["language"]
            if language != language_code:
                continue

            voice = Voice(
                name=server_voice["id"],
                gender=server_voice["gender"],
                region=server_voice["region"],
            )
            voices.append(voice)

        logging.debug(
            f"text_to_speech_api.get_available_voices: {voices} for language {language_code}"
        )

        return voices

    def _does_voice_supports_speeds(self):
        return False

    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        speed: float,
    ) -> str:

        url = urljoin(self.server, "/speak")
        url = f"{url}?voice={assigned_voice}&text={text}"
        response = requests.get(url)

        temp_filename = None
        with tempfile.NamedTemporaryFile(delete=False) as temporary_file:
            temp_filename = temporary_file.name

            if response.status_code == 200:
                with open(temp_filename, "wb") as f:
                    f.write(response.content)
            else:
                logging.error(
                    f"Failed to download the file. Status code: {response.status_code}"
                )

            self._convert_to_mp3(temp_filename, output_filename)

        logging.debug(
            f"text_to_speech_api._convert_text_to_speech: assigned_voice: {assigned_voice}, output_filename: '{output_filename}'"
        )
        return output_filename

    def get_languages(self):
        languages = set()
        for server_voice in self._get_voices():
            language = server_voice["language"]
            languages.add(language)

        languages = sorted(list(languages))
        logging.debug(f"text_to_speech_api.get_languages: {languages}")
        return languages



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings

from typing import Final

from moviepy import AudioFileClip, VideoFileClip, concatenate_videoclips

_DEFAULT_FPS: Final[int] = 30
_DEFAULT_DUBBED_VIDEO_FILE: Final[str] = "dubbed_video"
_DEFAULT_OUTPUT_FORMAT: Final[str] = ".mp4"


class VideoProcessing:

    @staticmethod
    def split_audio_video(*, video_file: str, output_directory: str) -> tuple[str, str]:
        """Splits an audio/video file into separate audio and video files."""

        base_filename = os.path.basename(video_file)
        filename, _ = os.path.splitext(base_filename)
        with VideoFileClip(video_file) as video_clip, warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            audio_clip = video_clip.audio
            audio_output_file = os.path.join(output_directory, filename + "_audio.mp3")
            audio_clip.write_audiofile(audio_output_file, logger=None)
            video_clip_without_audio = video_clip.with_audio(None)
            fps = video_clip.fps or _DEFAULT_FPS

            video_output_file = os.path.join(output_directory, filename + "_video.mp4")
            video_clip_without_audio.write_videofile(
                video_output_file, codec="libx264", fps=fps, logger=None
            )
        return video_output_file, audio_output_file

    @staticmethod
    def combine_audio_video(
        *,
        video_file: str,
        dubbed_audio_file: str,
        output_directory: str,
        target_language: str,
    ) -> str:
        """Combines an audio file with a video file, ensuring they have the same duration.

        Returns:
          The path to the output video file with dubbed audio.
        """

        video = VideoFileClip(video_file)
        audio = AudioFileClip(dubbed_audio_file)
        duration_difference = video.duration - audio.duration
        if duration_difference > 0:
            silence = AudioFileClip(duration=duration_difference).set_duration(
                duration_difference
            )
            audio = concatenate_videoclips([audio, silence])
        elif duration_difference < 0:
            audio = audio.subclipped(0, video.duration)
        final_clip = video.with_audio(audio)
        target_language_suffix = "_" + target_language.replace("-", "_").lower()
        dubbed_video_file = os.path.join(
            output_directory,
            _DEFAULT_DUBBED_VIDEO_FILE
            + target_language_suffix
            + _DEFAULT_OUTPUT_FORMAT,
        )
        final_clip.write_videofile(
            dubbed_video_file,
            codec="libx264",
            audio_codec="aac",
            temp_audiofile="temp-audio.m4a",
            remove_temp=True,
            logger=None,
        )
        return dubbed_video_file



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from pydub import AudioSegment
from transformers import Wav2Vec2Processor
from transformers.models.wav2vec2.modeling_wav2vec2 import (
    Wav2Vec2Model,
    Wav2Vec2PreTrainedModel,
)


class ModelHead(nn.Module):

    def __init__(self, config, num_labels):
        super().__init__()

        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = nn.Dropout(config.final_dropout)
        self.out_proj = nn.Linear(config.hidden_size, num_labels)

    def forward(self, features, **kwargs):

        x = features
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)

        return x


class AgeGenderModel(Wav2Vec2PreTrainedModel):

    def __init__(self, config):

        super().__init__(config)

        self.config = config
        self.wav2vec2 = Wav2Vec2Model(config)
        self.age = ModelHead(config, 1)
        self.gender = ModelHead(
            config, 3
        )  # Original model has 3 classes, we'll handle 2
        self.init_weights()

    def forward(
        self,
        input_values,
    ):

        outputs = self.wav2vec2(input_values)
        hidden_states = outputs[0]
        hidden_states = torch.mean(hidden_states, dim=1)
        logits_age = self.age(hidden_states)
        logits_gender = torch.softmax(self.gender(hidden_states), dim=1)

        return hidden_states, logits_age, logits_gender


class VoiceGenderClassifier:

    FEMALE = "Female"
    MALE = "Male"

    def __init__(self, device="cpu"):
        self.device = device
        # Load model from hub
        model_name = "audeering/wav2vec2-large-robust-24-ft-age-gender"
        self.processor = Wav2Vec2Processor.from_pretrained(model_name)
        self.model = AgeGenderModel.from_pretrained(model_name).to(self.device)

    # Function to load and process the MP3 file using pydub
    def load_audio_file(self, file_path, target_sampling_rate=16000):
        max_duration = 10  # Max duration in seconds

        # Load the audio file using pydub
        audio = AudioSegment.from_file(file_path, format="mp3")

        # If audio is longer than max_duration, trim it
        if len(audio) > max_duration * 1000:  # Convert seconds to milliseconds
            audio = audio[: max_duration * 1000]

        # Resample the audio to target sampling rate
        audio = audio.set_frame_rate(target_sampling_rate)

        # Convert to numpy array (pydub uses samples in the form of raw bytes)
        samples = np.array(audio.get_array_of_samples())

        # Ensure mono audio (convert stereo to mono if needed)
        if audio.channels == 2:
            samples = samples.reshape((-1, 2)).mean(axis=1)  # Average the two channels

        # Normalize the audio samples to the range [-1, 1] (convert to float32)
        samples = samples.astype(np.float32) / np.iinfo(np.int16).max

        # Ensure audio is in the correct shape (1, length)
        samples = np.expand_dims(samples, axis=0)

        return samples, target_sampling_rate

    def _predict(
        self,
        x: np.ndarray,
        sampling_rate: int,
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""Predict age and gender from raw audio signal."""

        # Run through processor to normalize signal
        y = self.processor(x, sampling_rate=sampling_rate)
        y = y["input_values"][0]
        y = y.reshape(1, -1)
        y = torch.from_numpy(y).to(self.device)

        # Run through model
        with torch.no_grad():
            y = self.model(y)
            logits_age, logits_gender = y[1], y[2]  # Age and gender logits
            return logits_age, logits_gender

    def _interpret_gender(self, logits_gender):
        """Convert gender logits into a male/female label."""
        # We're only interested in male and female, so we take the first two logits (ignore "child")
        male_female_logits = logits_gender[
            :, :2
        ]  # Take the first two logits (female, male)

        # Gender labels with only male and female
        gender_labels = [self.FEMALE, self.MALE]

        # Find the index with the highest probability between female and male
        predicted_gender_idx = torch.argmax(male_female_logits, dim=1).item()
        probabilities = F.softmax(male_female_logits, dim=1).cpu()
        prob_female, prob_male = probabilities[0].tolist()  # Since it's batch size of 1

        return gender_labels[predicted_gender_idx]

    def get_gender_for_file(self, file_path):
        try:
            signal, sampling_rate = self.load_audio_file(file_path)

            # Predict age and gender
            _, gender_logits = self._predict(signal, sampling_rate)
            predicted_gender = self._interpret_gender(gender_logits)
        except Exception as e:
            logging.error(
                f"voice_gender_classifier. get_gender_for_file. Error '{e}' processing {file_path}"
            )
            predicted_gender = self.FEMALE

        logging.debug(
            f"The audio from {os.path.basename(file_path)} is predicted {predicted_gender}"
        )
        return predicted_gender



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from typing import List

import numpy as np
import scipy.io.wavfile
import torch

from transformers import AutoTokenizer, VitsModel

from open_dubbing.text_to_speech import TextToSpeech, Voice


class TextToSpeechMMS(TextToSpeech):

    def __init__(self, device="cpu"):
        super().__init__()
        self.device = device
        logging.getLogger("transformers").setLevel(logging.ERROR)

    def get_available_voices(self, language_code: str) -> List[Voice]:
        return []

    def _convert_text_to_speech(
        self,
        *,
        assigned_voice: str,
        target_language: str,
        output_filename: str,
        text: str,
        speed: float,
    ) -> str:

        logging.debug(f"TextToSpeechMMS._convert_text_to_speech: {text}")
        local_files_only = False

        # Load pre-trained model and tokenizer
        model = VitsModel.from_pretrained(
            f"facebook/mms-tts-{target_language}", local_files_only=local_files_only
        ).to(self.device)
        tokenizer = AutoTokenizer.from_pretrained(
            f"facebook/mms-tts-{target_language}", local_files_only=local_files_only
        )
        inputs = tokenizer(text, return_tensors="pt").to(self.device)

        # Generate waveform
        with torch.no_grad():
            output = model(**inputs).waveform

        # Convert waveform to NumPy array and scale to 16-bit PCM
        # Assuming `output` is a 2D tensor with shape (batch_size, samples)
        output_np = output.squeeze().cpu().numpy()  # Remove batch dimension if present
        output_np = np.clip(output_np, -1, 1)  # Clip values to be between -1 and 1
        output_np = (output_np * 32767).astype(np.int16)  # Scale to 16-bit PCM

        # Get the sampling rate
        sampling_rate = model.config.sampling_rate

        # Write to WAV file
        wav_file = output_filename.replace(".mp3", ".wav")
        scipy.io.wavfile.write(wav_file, rate=sampling_rate, data=output_np)

        self._convert_to_mp3(wav_file, output_filename)
        logging.debug(
            f"text_to_speech.client.synthesize_speech: output_filename: '{output_filename}'"
        )
        return output_filename

    # Reference: https://dl.fbaipublicfiles.com/mms/tts/all-tts-languages.html
    def get_languages(self):
        return [
            "abi",
            "abp",
            "aca",
            "acd",
            "ace",
            "acf",
            "ach",
            "acn",
            "acr",
            "acu",
            "ade",
            "adh",
            "adj",
            "adx",
            "aeu",
            "agd",
            "agg",
            "agn",
            "agr",
            "agu",
            "agx",
            "aha",
            "ahk",
            "aia",
            "aka",
            "akb",
            "ake",
            "akp",
            "alj",
            "alp",
            "alt",
            "alz",
            "ame",
            "amf",
            "amh",
            "ami",
            "amk",
            "ann",
            "any",
            "aoz",
            "apb",
            "apr",
            "ara",
            "arl",
            "asa",
            "asg",
            "asm",
            "ata",
            "atb",
            "atg",
            "ati",
            "atq",
            "ava",
            "avn",
            "avu",
            "awa",
            "awb",
            "ayo",
            "ayr",
            "ayz",
            "azb",
            "azg",
            "azz",
            "bak",
            "bam",
            "ban",
            "bao",
            "bav",
            "bba",
            "bbb",
            "bbc",
            "bbo",
            "bcl",
            "bcw",
            "bdg",
            "bdh",
            "bdq",
            "bdu",
            "bdv",
            "beh",
            "bem",
            "ben",
            "bep",
            "bex",
            "bfa",
            "bfo",
            "bfy",
            "bfz",
            "bgc",
            "bgq",
            "bgr",
            "bgt",
            "bgw",
            "bha",
            "bht",
            "bhz",
            "bib",
            "bim",
            "bis",
            "biv",
            "bjr",
            "bjv",
            "bjw",
            "bjz",
            "bkd",
            "bkv",
            "blh",
            "blt",
            "blx",
            "blz",
            "bmq",
            "bmr",
            "bmu",
            "bmv",
            "bng",
            "bno",
            "bnp",
            "boa",
            "bod",
            "boj",
            "bom",
            "bor",
            "bov",
            "box",
            "bpr",
            "bps",
            "bqc",
            "bqi",
            "bqj",
            "bqp",
            "bru",
            "bsc",
            "bsq",
            "bss",
            "btd",
            "bts",
            "btt",
            "btx",
            "bud",
            "bul",
            "bus",
            "bvc",
            "bvz",
            "bwq",
            "bwu",
            "byr",
            "bzh",
            "bzi",
            "bzj",
            "caa",
            "cab",
            "cap",
            "car",
            "cas",
            "cat",
            "cax",
            "cbc",
            "cbi",
            "cbr",
            "cbs",
            "cbt",
            "cbu",
            "cbv",
            "cce",
            "cco",
            "cdj",
            "ceb",
            "ceg",
            "cek",
            "cfm",
            "cgc",
            "che",
            "chf",
            "chv",
            "chz",
            "cjo",
            "cjp",
            "cjs",
            "cko",
            "ckt",
            "cla",
            "cle",
            "cly",
            "cme",
            "cmr",
            "cnh",
            "cni",
            "cnl",
            "cnt",
            "coe",
            "cof",
            "cok",
            "con",
            "cot",
            "cou",
            "cpa",
            "cpb",
            "cpu",
            "crh",
            "crn",
            "crq",
            "crs",
            "crt",
            "csk",
            "cso",
            "ctd",
            "ctg",
            "cto",
            "ctu",
            "cuc",
            "cui",
            "cuk",
            "cul",
            "cwa",
            "cwe",
            "cwt",
            "cya",
            "cym",
            "daa",
            "dah",
            "dar",
            "dbj",
            "dbq",
            "ddn",
            "ded",
            "des",
            "deu",
            "dga",
            "dgi",
            "dgk",
            "dgo",
            "dgr",
            "dhi",
            "did",
            "dig",
            "dik",
            "dip",
            "div",
            "djk",
            "dnt",
            "dnw",
            "dop",
            "dos",
            "dsh",
            "dso",
            "dtp",
            "dts",
            "dug",
            "dwr",
            "dyi",
            "dyo",
            "dyu",
            "dzo",
            "eip",
            "eka",
            "ell",
            "emp",
            "enb",
            "eng",
            "enx",
            "ese",
            "ess",
            "eus",
            "evn",
            "ewe",
            "eza",
            "fal",
            "fao",
            "far",
            "fas",
            "fij",
            "fin",
            "flr",
            "fmu",
            "fon",
            "fra",
            "frd",
            "ful",
            "gai",
            "gam",
            "gau",
            "gbi",
            "gbk",
            "gbm",
            "gbo",
            "gde",
            "geb",
            "gej",
            "gil",
            "gjn",
            "gkn",
            "gld",
            "glk",
            "gmv",
            "gna",
            "gnd",
            "gng",
            "gog",
            "gor",
            "gqr",
            "grc",
            "gri",
            "grn",
            "grt",
            "gso",
            "gub",
            "guc",
            "gud",
            "guh",
            "guj",
            "guk",
            "gum",
            "guo",
            "guq",
            "guu",
            "gux",
            "gvc",
            "gvl",
            "gwi",
            "gwr",
            "gym",
            "gyr",
            "had",
            "hag",
            "hak",
            "hap",
            "hat",
            "hau",
            "hay",
            "heb",
            "heh",
            "hif",
            "hig",
            "hil",
            "hin",
            "hlb",
            "hlt",
            "hne",
            "hnn",
            "hns",
            "hoc",
            "hoy",
            "hto",
            "hub",
            "hui",
            "hun",
            "huu",
            "huv",
            "hvn",
            "hwc",
            "hyw",
            "iba",
            "icr",
            "idd",
            "ifa",
            "ifb",
            "ife",
            "ifk",
            "ifu",
            "ify",
            "ign",
            "ikk",
            "ilb",
            "ilo",
            "imo",
            "inb",
            "ind",
            "iou",
            "ipi",
            "iqw",
            "iri",
            "irk",
            "isl",
            "itl",
            "itv",
            "izr",
            "izz",
            "jac",
            "jam",
            "jav",
            "jbu",
            "jen",
            "jic",
            "jiv",
            "jmc",
            "jmd",
            "jun",
            "juy",
            "jvn",
            "kaa",
            "kab",
            "kac",
            "kak",
            "kan",
            "kao",
            "kaq",
            "kay",
            "kaz",
            "kbo",
            "kbp",
            "kbq",
            "kbr",
            "kby",
            "kca",
            "kcg",
            "kdc",
            "kde",
            "kdh",
            "kdi",
            "kdj",
            "kdl",
            "kdn",
            "kdt",
            "kek",
            "ken",
            "keo",
            "ker",
            "key",
            "kez",
            "kfb",
            "kfw",
            "kfx",
            "khg",
            "khm",
            "khq",
            "kia",
            "kij",
            "kik",
            "kin",
            "kir",
            "kjb",
            "kje",
            "kjg",
            "kjh",
            "kki",
            "kkj",
            "kle",
            "klu",
            "klv",
            "klw",
            "kma",
            "kmd",
            "kml",
            "kmu",
            "knb",
            "kne",
            "knf",
            "knj",
            "knk",
            "kno",
            "kog",
            "kor",
            "kpq",
            "kps",
            "kpv",
            "kpy",
            "kpz",
            "kqe",
            "kqp",
            "kqr",
            "kqy",
            "krc",
            "kri",
            "krj",
            "krl",
            "krr",
            "kru",
            "ksb",
            "ksr",
            "kss",
            "ktb",
            "ktj",
            "kub",
            "kue",
            "kum",
            "kus",
            "kvn",
            "kvw",
            "kwd",
            "kwf",
            "kwi",
            "kxc",
            "kxf",
            "kxm",
            "kxv",
            "kyb",
            "kyc",
            "kyf",
            "kyg",
            "kyo",
            "kyq",
            "kyu",
            "kyz",
            "kzf",
            "lac",
            "laj",
            "lam",
            "lao",
            "las",
            "lat",
            "lav",
            "law",
            "lbj",
            "lbw",
            "lcp",
            "lee",
            "lef",
            "lem",
            "lew",
            "lex",
            "lgg",
            "lgl",
            "lhu",
            "lia",
            "lid",
            "lif",
            "lip",
            "lis",
            "lje",
            "ljp",
            "llg",
            "lln",
            "lme",
            "lnd",
            "lns",
            "lob",
            "lok",
            "lom",
            "lon",
            "loq",
            "lsi",
            "lsm",
            "luc",
            "lug",
            "lwo",
            "lww",
            "lzz",
            "mad",
            "mag",
            "mah",
            "mai",
            "maj",
            "mak",
            "mal",
            "maq",
            "mar",
            "maw",
            "maz",
            "mbb",
            "mbc",
            "mbh",
            "mbj",
            "mbt",
            "mbu",
            "mbz",
            "mca",
            "mcb",
            "mcd",
            "mco",
            "mcp",
            "mcq",
            "mcu",
            "mda",
            "mdv",
            "mdy",
            "med",
            "mee",
            "mej",
            "men",
            "meq",
            "met",
            "mev",
            "mfe",
            "mfh",
            "mfi",
            "mfk",
            "mfq",
            "mfy",
            "mfz",
            "mgd",
            "mge",
            "mgh",
            "mgo",
            "mhi",
            "mhr",
            "mhu",
            "mhx",
            "mhy",
            "mib",
            "mie",
            "mif",
            "mih",
            "mil",
            "mim",
            "min",
            "mio",
            "mip",
            "miq",
            "mit",
            "miy",
            "miz",
            "mjl",
            "mjv",
            "mkl",
            "mkn",
            "mlg",
            "mmg",
            "mnb",
            "mnf",
            "mnk",
            "mnw",
            "mnx",
            "moa",
            "mog",
            "mon",
            "mop",
            "mor",
            "mos",
            "mox",
            "moz",
            "mpg",
            "mpm",
            "mpp",
            "mpx",
            "mqb",
            "mqf",
            "mqj",
            "mqn",
            "mrw",
            "msy",
            "mtd",
            "mtj",
            "mto",
            "muh",
            "mup",
            "mur",
            "muv",
            "muy",
            "mvp",
            "mwq",
            "mwv",
            "mxb",
            "mxq",
            "mxt",
            "mxv",
            "mya",
            "myb",
            "myk",
            "myl",
            "myv",
            "myx",
            "myy",
            "mza",
            "mzi",
            "mzj",
            "mzk",
            "mzm",
            "mzw",
            "nab",
            "nag",
            "nan",
            "nas",
            "naw",
            "nca",
            "nch",
            "ncj",
            "ncl",
            "ncu",
            "ndj",
            "ndp",
            "ndv",
            "ndy",
            "ndz",
            "neb",
            "new",
            "nfa",
            "nfr",
            "nga",
            "ngl",
            "ngp",
            "ngu",
            "nhe",
            "nhi",
            "nhu",
            "nhw",
            "nhx",
            "nhy",
            "nia",
            "nij",
            "nim",
            "nin",
            "nko",
            "nlc",
            "nld",
            "nlg",
            "nlk",
            "nmz",
            "nnb",
            "nnq",
            "nnw",
            "noa",
            "nod",
            "nog",
            "not",
            "npl",
            "npy",
            "nst",
            "nsu",
            "ntm",
            "ntr",
            "nuj",
            "nus",
            "nuz",
            "nwb",
            "nxq",
            "nya",
            "nyf",
            "nyn",
            "nyo",
            "nyy",
            "nzi",
            "obo",
            "oku",
            "old",
            "omw",
            "onb",
            "ood",
            "orm",
            "ory",
            "oss",
            "ote",
            "otq",
            "ozm",
            "pab",
            "pad",
            "pag",
            "pam",
            "pan",
            "pao",
            "pap",
            "pau",
            "pbb",
            "pbc",
            "pbi",
            "pce",
            "pcm",
            "peg",
            "pez",
            "pib",
            "pil",
            "pir",
            "pis",
            "pjt",
            "pkb",
            "pls",
            "plw",
            "pmf",
            "pny",
            "poi",
            "pol",
            "por",
            "poy",
            "ppk",
            "pps",
            "prf",
            "prk",
            "prt",
            "pse",
            "pss",
            "ptu",
            "pui",
            "pwg",
            "pww",
            "pxm",
            "qub",
            "quf",
            "quh",
            "qul",
            "quw",
            "quy",
            "quz",
            "qvc",
            "qve",
            "qvh",
            "qvm",
            "qvn",
            "qvo",
            "qvs",
            "qvw",
            "qvz",
            "qwh",
            "qxh",
            "qxl",
            "qxn",
            "qxo",
            "qxr",
            "rah",
            "rai",
            "rap",
            "rav",
            "raw",
            "rej",
            "rel",
            "rgu",
            "rhg",
            "ril",
            "rim",
            "rjs",
            "rkt",
            "rmo",
            "rng",
            "rnl",
            "rol",
            "ron",
            "rop",
            "rro",
            "rub",
            "ruf",
            "rug",
            "run",
            "rus",
            "sab",
            "sag",
            "sah",
            "saj",
            "saq",
            "sas",
            "sba",
            "sbd",
            "sbl",
            "sbp",
            "sch",
            "sck",
            "sda",
            "sea",
            "seh",
            "ses",
            "sey",
            "sgb",
            "sgj",
            "sgw",
            "shi",
            "shk",
            "shn",
            "sho",
            "shp",
            "sid",
            "sig",
            "sil",
            "sja",
            "sjm",
            "sld",
            "slu",
            "sml",
            "smo",
            "sna",
            "sne",
            "snn",
            "snp",
            "snw",
            "som",
            "soy",
            "spa",
            "spp",
            "spy",
            "sqi",
            "sri",
            "srm",
            "srn",
            "srx",
            "stn",
            "stp",
            "suc",
            "suk",
            "sun",
            "sur",
            "sus",
            "suv",
            "suz",
            "swe",
            "swh",
            "sxb",
            "sxn",
            "sya",
            "syl",
            "sza",
            "tac",
            "taj",
            "tam",
            "tao",
            "tap",
            "taq",
            "tat",
            "tav",
            "tbc",
            "tbg",
            "tbk",
            "tbl",
            "tby",
            "tbz",
            "tca",
            "tcc",
            "tcs",
            "tcz",
            "tdj",
            "ted",
            "tee",
            "tel",
            "tem",
            "teo",
            "ter",
            "tes",
            "tew",
            "tex",
            "tfr",
            "tgj",
            "tgk",
            "tgl",
            "tgo",
            "tgp",
            "tha",
            "thk",
            "thl",
            "tih",
            "tik",
            "tir",
            "tkr",
            "tlb",
            "tlj",
            "tly",
            "tmc",
            "tmf",
            "tna",
            "tng",
            "tnk",
            "tnn",
            "tnp",
            "tnr",
            "tnt",
            "tob",
            "toc",
            "toh",
            "tom",
            "tos",
            "tpi",
            "tpm",
            "tpp",
            "tpt",
            "trc",
            "tri",
            "trn",
            "trs",
            "tso",
            "tsz",
            "ttc",
            "tte",
            "tue",
            "tuf",
            "tuo",
            "tur",
            "tvw",
            "twb",
            "twe",
            "twu",
            "txa",
            "txq",
            "txu",
            "tye",
            "udm",
            "udu",
            "ukr",
            "unr",
            "upv",
            "ura",
            "urb",
            "urk",
            "urt",
            "ury",
            "usp",
            "vag",
            "vid",
            "vie",
            "vif",
            "vmw",
            "vmy",
            "vun",
            "vut",
            "wap",
            "war",
            "waw",
            "way",
            "wba",
            "wlo",
            "wlx",
            "wmw",
            "wob",
            "wsg",
            "wwa",
            "xal",
            "xdy",
            "xed",
            "xer",
            "xmm",
            "xnj",
            "xnr",
            "xog",
            "xon",
            "xrb",
            "xsb",
            "xsm",
            "xsr",
            "xsu",
            "xta",
            "xtd",
            "xte",
            "xtm",
            "xtn",
            "xua",
            "xuo",
            "yaa",
            "yad",
            "yal",
            "yam",
            "yao",
            "yas",
            "yat",
            "yaz",
            "yba",
            "ybb",
            "ycl",
            "ycn",
            "yea",
            "yka",
            "yli",
            "yor",
            "yre",
            "yua",
            "yuz",
            "yva",
            "zaa",
            "zab",
            "zac",
            "zad",
            "zae",
            "zai",
            "zam",
            "zao",
            "zaq",
            "zar",
            "zas",
            "zav",
            "zaw",
            "zca",
            "zga",
            "zim",
            "ziw",
            "zlm",
            "zmz",
            "zne",
            "zos",
            "zpc",
            "zpg",
            "zpi",
            "zpl",
            "zpm",
            "zpo",
            "zpt",
            "zpu",
            "zpz",
            "ztq",
            "zty",
            "zyb",
            "zyp",
            "zza",
        ]



# Copyright 2024 Jordi Mas i Hernàndez <jmas@softcatala.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import logging

import numpy as np
import torch

from pydub import AudioSegment
from transformers import WhisperForConditionalGeneration, WhisperProcessor

from open_dubbing.speech_to_text import SpeechToText


class SpeechToTextWhisperTransformers(SpeechToText):

    def __init__(self, *, model_name="medium", device="cpu", cpu_threads=0):
        super().__init__(device=device, model_name=model_name, cpu_threads=cpu_threads)
        self._processor = None

    def load_model(self):
        full_model_name = f"openai/whisper-{self.model_name}"
        self._processor = WhisperProcessor.from_pretrained(full_model_name)
        self._model = WhisperForConditionalGeneration.from_pretrained(full_model_name)

    def _transcribe(
        self,
        *,
        vocals_filepath: str,
        source_language_iso_639_1: str,
    ) -> str:

        audio = AudioSegment.from_file(vocals_filepath)
        audio = audio.set_channels(1)  # Convert to mono
        audio = audio.set_frame_rate(16000)  # Set the frame rate to 16kHz
        # Convert the audio to a numpy array
        audio_input = (
            np.array(audio.get_array_of_samples()).astype(np.float32) / 32768.0
        )  # Normalize

        # Preprocess the audio input
        input_features = self._processor(
            audio_input, sampling_rate=16000, return_tensors="pt"
        ).input_features

        with torch.no_grad():
            generated_ids = self._model.generate(
                input_features, language=source_language_iso_639_1
            )
        transcription = self._processor.batch_decode(
            generated_ids, skip_special_tokens=True
        )[0]
        logging.debug(
            f"speech_to_text_whisper_transfomers._transcribe. transcription: {transcription}, file {vocals_filepath}"
        )
        return transcription

    def _get_audio_language(self, audio: array.array) -> str:
        audio_input = np.array(audio).astype(np.float32) / 32768.0

        # Preprocess the audio input
        input_features = self._processor(
            audio_input, sampling_rate=16000, return_tensors="pt"
        ).input_features

        with torch.no_grad():
            generated_ids = self._model.generate(input_features)

        # Decode the transcription including special tokens to capture the language token
        transcription_with_tokens = self._processor.batch_decode(
            generated_ids, skip_special_tokens=False
        )[0]

        detected_language = None
        if "|" in transcription_with_tokens:
            for token in transcription_with_tokens.split("|"):
                if (len(token) == 2 or len(token) == 3) and token.isalpha():
                    detected_language = self._get_iso_639_3(token)
                    break
        logging.debug(
            f"speech_to_text_whisper_transfomers._get_audio_language. Detected language: {detected_language}"
        )

        return detected_language

    def get_languages(self):
        languages = [
            "af",
            "am",
            "ar",
            "as",
            "az",
            "ba",
            "be",
            "bg",
            "bn",
            "bo",
            "br",
            "bs",
            "ca",
            "cs",
            "cy",
            "da",
            "de",
            "el",
            "en",
            "es",
            "et",
            "eu",
            "fa",
            "fi",
            "fo",
            "fr",
            "gl",
            "gu",
            "ha",
            "haw",
            "he",
            "hi",
            "hr",
            "ht",
            "hu",
            "hy",
            "id",
            "is",
            "it",
            "ja",
            "jw",
            "ka",
            "kk",
            "km",
            "kn",
            "ko",
            "la",
            "lb",
            "ln",
            "lo",
            "lt",
            "lv",
            "mg",
            "mi",
            "mk",
            "ml",
            "mn",
            "mr",
            "ms",
            "mt",
            "my",
            "ne",
            "nl",
            "nn",
            "no",
            "oc",
            "pa",
            "pl",
            "ps",
            "pt",
            "ro",
            "ru",
            "sa",
            "sd",
            "si",
            "sk",
            "sl",
            "sn",
            "so",
            "sq",
            "sr",
            "su",
            "sv",
            "sw",
            "ta",
            "te",
            "tg",
            "th",
            "tk",
            "tl",
            "tr",
            "tt",
            "uk",
            "ur",
            "uz",
            "vi",
            "yi",
            "yo",
            "zh",
            "yue",
        ]

        iso_639_3 = []
        for language in languages:
            pt3 = self._get_iso_639_3(language)
            iso_639_3.append(pt3)
        return iso_639_3



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import subprocess
import sys


class Demucs:

    def build_demucs_command(
        self,
        *,
        audio_file: str,
        output_directory: str,
        device: str = "cpu",
        shifts: int = 1,
        overlap: float = 0.25,
        mp3_bitrate: int = 320,
        mp3_preset: int = 2,
        jobs: int = 0,
        split: bool = True,
        segment: int | None = None,
        mp3: bool = True,
    ) -> str:
        """Builds the Demucs audio separation command.

        Args:
            audio_file: The path to the audio file to process.
            output_directory: The output directory for separated tracks.
            device: The device to use ("cuda" or "cpu").
            shifts: The number of random shifts for equivariant stabilization.
            overlap: The overlap between splits.
            mp3_bitrate: The bitrate of converted MP3 files.
            mp3_preset: The encoder preset for MP3 conversion.
            jobs: The number of jobs to run in parallel.
            split: Whether to split audio into chunks.
            segment: The split size for chunks (None for no splitting).
            mp3: Convert output to MP3.

        Returns:
            A string representing the constructed command.

        """
        command_parts = [
            sys.executable,
            "-m",
            "demucs.separate",
            "-o",
            f'"{output_directory}"',
            "--device",
            device,
            "--shifts",
            str(shifts),
            "--overlap",
            str(overlap),
            "-j",
            str(jobs),
            "--two-stems vocals",
        ]
        if not split:
            command_parts.append("--no-split")
        elif segment is not None:
            command_parts.extend(["--segment", str(segment)])
        if mp3:
            command_parts.extend(
                [
                    "--mp3",
                    "--mp3-bitrate",
                    str(mp3_bitrate),
                    "--mp3-preset",
                    str(mp3_preset),
                ]
            )
        command_parts.append(f'"{audio_file}"')
        return " ".join(command_parts)

    def execute_demucs_command(self, command: str) -> None:
        """Executes a Demucs command using subprocess.

        Demucs is a model using AI/ML to detach dialogues
        from the rest of the audio file.

        Args:
            command: The string representing the command to execute.
        """
        try:
            result = subprocess.run(
                command, shell=True, capture_output=True, text=True, check=True
            )
            logging.debug(result.stdout)
        except subprocess.CalledProcessError as error:
            raise Exception(
                f"Error in attempt to separate audio: {error}\n{error.stderr}"
            )

    def _extract_command_info(self, command: str) -> tuple[str, str, str]:
        """Extracts folder name, output file extension, and input file name (without path) from a Demucs command.

        Args:
            command: The Demucs command string.

        Returns:
            tuple: A tuple containing (folder_name, output_file_extension,
            input_file_name).
        """
        folder_pattern = r"-o\s+(['\"]?)(.+?)\1"
        mp3_pattern = r"--mp3"
        input_file_pattern = r"['\"]?(\w+\.\w+)['\"]?$|\s(\w+\.\w+)$"
        folder_match = re.search(folder_pattern, command)
        mp3_match = re.search(mp3_pattern, command)
        input_file_match = re.search(input_file_pattern, command)
        output_directory = folder_match.group(2) if folder_match else ""
        input_file_name_with_ext = (
            (input_file_match.group(1) or input_file_match.group(2))
            if input_file_match
            else ""
        )
        input_file_name_no_ext = (
            os.path.splitext(input_file_name_with_ext)[0] if input_file_match else ""
        )
        if mp3_match:
            output_file_extension = ".mp3"
        else:
            output_file_extension = ".wav"
        return output_directory, output_file_extension, input_file_name_no_ext

    def assemble_split_audio_file_paths(self, command: str) -> tuple[str, str]:
        """Returns paths to the audio files with vocals and no vocals.

        Returns:
            A tuple with a path to the file with the audio with vocals only
            and the other with the background sound only.
        """
        MODEL_NAME = "htdemucs"
        output_directory, output_file_extension, input_file_name = (
            self._extract_command_info(command)
        )
        audio_vocals_file = os.path.join(
            output_directory,
            MODEL_NAME,
            input_file_name,
            f"vocals{output_file_extension}",
        )
        audio_background_file = os.path.join(
            output_directory,
            MODEL_NAME,
            input_file_name,
            f"no_vocals{output_file_extension}",
        )
        return audio_vocals_file, audio_background_file
